import numpy as np
import os
import csv

class Bin(object):
    ''' Container for items that keeps a running sum '''
    def __init__(self):
        self.items = []
        self.sum = 0

    def append(self, item):
        self.items.append(item)
        self.sum += item

    def __str__(self):
        ''' Printable representation '''
        return 'Bin(sum=%d, items=%s)' % (self.sum, str(self.items))

# best fit decreasing
def BFDpack(weights, C):
    bins = []

    for item in weights:
        if len(bins) == 0:
            bin = Bin()
            bin.append(item)
            bins.append(bin)
        else:

            min_waste = [C-item-bin.sum for bin in bins]

            for i in range(len(min_waste)):
                if min_waste[i]<0:
                    min_waste[i] += 10*C # ensure this bin is not chosen if possible
            min_waste_index = np.argmin(min_waste)
            if min_waste[min_waste_index]>5*C:
                # item didn't fit into any bin, start a new bin
                #print 'Making new bin for', item
                bin = Bin()
                bin.append(item)
                bins.append(bin)
            else:
                bins[min_waste_index].append(item)

    return bins

# first fit decreasing
def FFDpack(weights, C):
    bins = []

    for item in weights:
        # Try to fit item into a bin

        for bin in bins:
            if bin.sum + item <= C:
                #print 'Adding', item, 'to', bin
                bin.append(item)
                break
        else:
            # item didn't fit into any bin, start a new bin
            #print 'Making new bin for', item
            bin = Bin()
            bin.append(item)
            bins.append(bin)

    return bins

# Djang and Finch algorithm
# C/m is the threshold before we try to fill the bin, typically m=3
# K is how much we increase the tolerance after failing to fill a bin
# often K=C/20, 1 is an alternative
def DJDpack(weights, C, m, K):

    if K<1:
        # might as well round K up to 1
        K=1

    bins = []
    w = []
    w[:] = weights

    while len(w)>0:
        # fill one bin at a time
        bin = Bin()
        # fill the bin up to C/m
        while bin.sum < C/m and len(w) > 0:
            bin.append(w[0])
            w.pop(0)
        if len(w) == 0:
            # no weights left, so end process
            bins.append(bin)
            return bins

        elif bin.sum + w[-1] > C:
            # can't fit any items into the bin, so don't do anything
            1

        else:
            # fill the bin with up to triplets according to DJD procedure
            tolerance = 0
            while True:
                combo = DJDsubpack(w, C, bin.sum, tolerance)
                if combo:
                    # we have a valid combination
                    combo_weights = [w[i] for i in combo]
                    combo_weights.reverse()
                    for index in combo:
                        w.pop(index)
                    for weight in combo_weights:
                        bin.append(weight)
                    break
                tolerance += K

        bins.append(bin)
    return bins

# a function that returns a tuple consisting of a valid combination, and empty list
# largest index/smallest weight is first
def DJDsubpack(w, C, current_weight, tolerance):
    if w[0]+current_weight >= C-tolerance and w[-1]+current_weight <= C:
        # try 1-item combinations
            for i in range(len(w)):
                if C-tolerance <= current_weight+w[i] <= C:
                    return [i]

    if len(w) >= 2:
        if w[0]+w[1]+current_weight >= C-tolerance and w[-1]+w[-2]+current_weight <= C:
            # try 2-item combinations
            for i in range(1, len(w)):
                for j in range(i):
                    if C-tolerance <= current_weight+w[i]+w[j] <= C:
                        return [i, j]

    if len(w) >= 3:
        if w[0]+w[1]+w[2]+current_weight >= C-tolerance and w[-1]+w[-2]+w[-3]+current_weight <= C:
            for i in range(2, len(w)):
                for j in range(1, i):
                    for k in range(j):
                        if C-tolerance <= current_weight+w[i]+w[j]+w[k] <= C:
                            return [i, j, k]

    # no valid combination
    return []
